
# THIS FILE IS GENERATED FROM ORANGE SETUP.PY
short_version = '3.24.0'
version = '3.24.0'
full_version = '3.24.0'
git_revision = 'Unknown'
release = True

if not release:
    version = full_version
    short_version += ".dev"
